#!/usr/bin/env python3

import os
import socketio
import logging
import requests
import re
import time


CYTUBE_CSRF_RE = re.compile(r'<input type="hidden" name="_csrf" value="([^"]+)">')


try:
    from instance import config
except ImportError:
    raise ImportError("Guntstream hasn't been configured yet! "
                      "Please place a configuration file at "
                      "instance/config.py.")

# PID for managing through uWSGI

pidfile = open(config.GUNTBOT_PIDFILE, "w")
pidfile.write(str(os.getpid()))
pidfile.close()

# Logging

logger = logging.getLogger(__name__)
if config.GUNTBOT_DEBUG:
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.INFO)
logging.basicConfig()

# Authentication

authdata = {
    "value": None,
    "expiry": None
}


def renew_auth(authdata):
    logger.info("Authentication does not exist. Trying to log in.")

    session = requests.session()
    # The cytu.be request requires a CSRF token to complete. Grab that.
    login_html = session.get("https://cytu.be/login").text
    tokens = CYTUBE_CSRF_RE.findall(login_html)
    if not tokens:
        logger.warning("Couldn't find any CSRF tokens on login page!")
        authdata["value"] = None
        return

    resp = session.post("https://cytu.be/login",
                        data={
                            "_csrf": tokens[0],
                            "name": config.GUNTBOT_USERNAME,
                            "password": config.GUNTBOT_PASSWORD,
                            "remember": "on"
                        },
                        allow_redirects=False)
    if resp.status_code != 200:
        logger.warning("Unsuccessful login.")
        authdata["value"] = None
        return

    found_cookie = False
    for cookie in list(session.cookies):
        if cookie.name == "auth":
            found_cookie = True
            authfile.truncate(0)
            authfile.write(cookie.value)
            authfile.write("\n")

            if cookie.expires is not None:
                expiry = cookie.expires
            else:
                expiry = (2**63)-1

            authfile.write(str(expiry))
            authfile.flush()

            authdata["value"] = cookie.value
            authdata["expiry"] = expiry
    if not found_cookie:
        logger.warning("No authentication cookie found.")
        authdata["value"] = None


try:
    authfile = open(config.GUNTBOT_AUTHFILE, "r+")
    authfile.seek(0)
    try:
        authdata["value"], authdata["expiry"] = authfile.read().splitlines()
        authfile.seek(0)

        if int(authdata["expiry"]) < time.time():
            renew_auth(authdata)
    except ValueError:
        renew_auth(authdata)
except FileNotFoundError:
    authfile = open(config.GUNTBOT_AUTHFILE, "w+")
    renew_auth(authdata)
authfile.close()


# Globals


users = None


# Commands


selected_stream = None
current_id = None


def handle_stream(args, user):
    global selected_stream
    if user["rank"] < 2:
        # Ignore.
        return

    selected_stream = args[0]

    sio.emit("queue", {
        "id": config.GUNTBOT_STREAMLINK_TEMPLATE.format(selected_stream),
        "type": "cm",
        "pos": "next",
        "temp": True
    })


def handle_remove(args, user):
    global selected_stream, current_id
    if user["rank"] < 2:
        # Ignore.
        return

    selected_stream = None

    if current_id is not None:
        sio.emit("delete", current_id)
        current_id = None
        msg = "Stream removed."
    else:
        msg = "No active streams, or I didn't add the current stream."

    sio.emit("chatMsg", {
        "msg": msg,
        "meta": {
            "modflair": find_user(config.GUNTBOT_USERNAME)["rank"]
        }
    })


# Message parsing


def find_user(username):
    if users is None:
        return None

    for user in users:
        if user["name"] == username:
            return user

    return None


def parse_message(message, user):
    if not message.startswith("!"):
        return

    user = find_user(user)
    if user is None:
        return

    args = message.split(" ")
    command = args.pop(0)[1:]

    command_func = "handle_{}".format(command)
    if command_func not in globals().keys():
        return

    globals()[command_func](args, user)


# Connection


sio = socketio.Client(logger=logger)


@sio.event
def connect():
    logger.info("Connected!")
    sio.emit("joinChannel", {
        "name": config.GUNTBOT_CHANNEL
    })


@sio.event
def rank(r):
    logger.info("My rank now set to %d", r)
    if r == -1:  # Anonymous
        logger.warning("Not logged in! Logging in as test user.")
        sio.emit("login", {
            "name": config.GUNTBOT_USERNAME + "_test"
        })
    elif r < 2:  # 2 = Moderator
        logger.warning("My rank is lower than moderator, "
                       "I may not be able to manage things!")


@sio.event
def userlist(data):
    global users
    users = data


@sio.event
def queueFail(data):
    if selected_stream is not None:
        sio.emit("chatMsg", {
            "msg": "Couldn't queue stream for {}, make sure they're live."
            .format(selected_stream),
            "meta": {
                "modflair": find_user(config.GUNTBOT_USERNAME)["rank"]
            }
        })


@sio.event
def queue(data):
    global current_id
    if data["item"]["queueby"] == config.GUNTBOT_USERNAME:
        # Switch to it
        sio.emit("jumpTo", data["item"]["uid"])
        current_id = data["item"]["uid"]
        sio.emit("chatMsg", {
            "msg": "Take it away, {}!".format(selected_stream),
            "meta": {
                "modflair": find_user(config.GUNTBOT_USERNAME)["rank"]
            }
        })


@sio.event
def chatMsg(data):
    parse_message(data["msg"], data["username"])


@sio.event
def addUser(data):
    if users is not None:
        users.append(data)


@sio.event
def userLeave(data):
    if users is not None:
        users.remove(find_user(data["name"]))


@sio.event
def disconnect():
    logger.info("Disconnected.")


if __name__ == "__main__":
    try:
        if authdata["value"] is not None:
            headers = {
                "Cookie": "auth={}".format(authdata["value"])
            }
        else:
            headers = {}
        sio.connect("https://vapor.cytu.be:10443/socket.io", headers=headers)
        sio.wait()
    except KeyboardInterrupt:
        sio.disconnect()
        logger.info("Shutting down the bot...")
