import uuid
import functools
from flask import (
    Blueprint, render_template, request, session, current_app,
    url_for, redirect, flash, g
)
from werkzeug.security import check_password_hash, generate_password_hash

from .db import db, Channel


bp = Blueprint('auth', __name__)


ERROR_NO_USERNAME = "Username is required."
ERROR_USERNAME_LENGTH = "Username must be between 2 and 32 characters."
ERROR_NO_PASSWORD = "Password is required."
ERROR_USER_EXISTS = "A channel with the same name already exists."
ERROR_MASTER_KEY_NOMATCH = "Master key is not valid."
SUCCESS = "Success! Your account was created."


@bp.route("/create", methods=("GET", "POST"))
def create_channel():
    if request.method == "POST":
        username = request.form['username']
        password = request.form['password']
        master_key = request.form['masterkey']

        if not username:
            error = ERROR_NO_USERNAME
        elif not 2 <= len(username) <= 32:
            error = ERROR_USERNAME_LENGTH
        elif not password:
            error = ERROR_NO_PASSWORD
        elif Channel.query.filter_by(username=username).count():
            error = ERROR_USER_EXISTS
        elif master_key != current_app.config["MASTER_KEY"]:
            error = ERROR_MASTER_KEY_NOMATCH
        else:
            error = None

        if error is None:
            channel = Channel(
                username=username, password=generate_password_hash(password),
                stream_key=str(uuid.uuid4())
            )
            db.session.add(channel)
            db.session.commit()
            flash(SUCCESS, "info")
            return redirect(url_for('auth.login'))

        flash(error, "error")
    return render_template("auth/register.html")


ERROR_NO_SUCH_USER = "User doesn't exist."
ERROR_INVALID_PASSWORD = "Invalid password."


def do_login():
    """The actual login logic."""
    username = request.form["username"]
    password = request.form["password"]

    if not username:
        return None, ERROR_NO_USERNAME
    elif not password:
        return None, ERROR_NO_PASSWORD

    channel = Channel.query.filter_by(username=username)
    if not channel.count():
        return None, ERROR_NO_SUCH_USER
    channel = channel.first()
    if not check_password_hash(channel.password, password):
        return None, ERROR_INVALID_PASSWORD

    session.clear()
    session["user_id"] = channel.id
    return redirect(url_for("channel.index")), None


@bp.route("/login", methods=("GET", "POST"))
def login():
    if request.method == "POST":
        resp, error = do_login()
        if error is None:
            return resp

        flash(error, "error")
    return render_template("auth/login.html")


@bp.before_app_request
def load_logged_in_user():
    user_id = session.get("user_id")

    if user_id is None:
        g.user = None
    else:
        g.user = Channel.query.get(user_id)


def login_required(view):
    @functools.wraps(view)
    def wrapped_view(*args, **kwargs):
        if g.user is None:
            return redirect(url_for('auth.login'))
        return view(*args, **kwargs)
    return wrapped_view


@bp.route("/logout", methods=("GET",))
@login_required
def logout():
    session.pop('user_id', None)
    return redirect(url_for('.login'))
