import uuid
import os
import os.path
import subprocess

from flask import (
    g, request, Blueprint, render_template, flash, jsonify,
    redirect, url_for, current_app, send_from_directory
)
from werkzeug.utils import secure_filename
from sqlalchemy import exists

from .auth import login_required
from .db import db, Emote, Channel


bp = Blueprint("channel", __name__)


def make_error(err, url, status_code=400):
    if request.args.get("ajax"):
        return jsonify({
            "error": err
        }), status_code
    else:
        flash(err, "error")
        return redirect(url)


@bp.route("/")
@login_required
def index():
    return render_template("channel/index.html")


ERROR_NO_TITLE = "Title cannot be empty."
SUCCESS_TITLE = "Title successfully updated."


@bp.route("/update_title", methods=("POST",))
@login_required
def update_title():
    title = request.form["title"]

    if not title:
        return make_error(ERROR_NO_TITLE, url_for("channel.index"))

    g.user.stream_title = title
    db.session.add(g.user)
    db.session.commit()

    if request.args.get("ajax"):
        return "", 204
    else:
        flash(SUCCESS_TITLE, "info")
        return redirect(url_for("channel.index"))


SUCCESS_STREAM_KEY_RESET = "Stream key successfully reset."


@bp.route("/reset_stream_key", methods=("POST",))
@login_required
def reset_stream_key():
    g.user.stream_key = str(uuid.uuid4())
    db.session.add(g.user)
    db.session.commit()

    if request.args.get("ajax"):
        return jsonify({
            "stream_key": g.user.stream_key
        }), 200
    else:
        flash(SUCCESS_STREAM_KEY_RESET, "info")
        return redirect(url_for("channel.index"))


@bp.route("/emotes/", methods=("GET",))
@bp.route("/emotes/<string:username>")
@login_required
def emotes(username: str = None):
    if username:
        emotes = Emote.query.join(Channel).filter(
            Channel.username == username,
            Emote.channel_id == Channel.id,
        )
    else:
        emotes = Emote.query.all()

    return render_template(
        "channel/emotes.html",
        emotes=emotes,
        all_emotes=not username,
        username=username
    )


ERROR_NO_FILES = "You didn't send any files."
ERROR_UNSUPPORTED_FILE = "Emote {} has an unsupported filetype."
ERROR_EMOTE_EXISTS = "Emote {} already exists."


def allowed_file(filename):
    return '.' in filename and \
        filename.rsplit('.', 1)[1].lower() in {
            'gif', 'png', 'jpg', 'jpeg'
        }


def resize_image(emote, ext):
    proc = subprocess.Popen([
        "convert", ext + ":-", "-resize", current_app.config["EMOTE_SIZE"],
        ext + ":-"
    ], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    data, _ = proc.communicate(emote.read())
    return data


@bp.route("/create_emote", methods=("POST",))
@login_required
def create_emote():
    return_url = request.headers.get("Referer", url_for(".emotes"))

    if "files" not in request.files or not request.files.getlist("files"):
        return make_error(ERROR_NO_FILES, return_url)

    created_emotes = []
    for file in request.files.getlist("files"):
        # Check extension
        if not allowed_file(file.filename):
            flash(
                ERROR_UNSUPPORTED_FILE.format(file.filename),
                "warning"
            )
            continue

        filename = secure_filename(file.filename)
        emote_name, ext = filename.rsplit(".", 1)

        # Check if emote name is unique
        if db.session.query(exists().where(
                Emote.emote_name == emote_name
        )).scalar():
            return make_error(
                ERROR_EMOTE_EXISTS.format(emote_name), return_url,
            )

        # Save emote in the filesystem
        if not filename or \
           os.path.exists(
               os.path.join(
                   current_app.config["UPLOAD_FOLDER"],
                   filename
               )
           ):
            filename = (
                str(uuid.uuid4()).replace("-", "")
                + "." + filename.rsplit(".", 1)[1]
            )

        converted_file = resize_image(file, ext)
        with open(os.path.join(
                current_app.config["UPLOAD_FOLDER"],
                filename
        ), "wb") as f:
            f.write(converted_file)

        # Create emote
        new_emote = Emote(
            channel_id=g.user.id,
            emote_name=emote_name,
            emote_url=filename
        )
        db.session.add(new_emote)
        db.session.commit()
        created_emotes.append(new_emote)

    if request.args.get("ajax"):
        return jsonify({
            "messages": {
                # TODO handle warnings
            },
            "emotes": [
                {
                    "name": emote.emote_name,
                    "url": os.path.join(
                        current_app.config["UPLOAD_FOLDER"],
                        emote.emote_url
                    )
                } for emote in created_emotes
            ]
        })
    else:
        flash("Emotes uploaded.", "info")
        return redirect(return_url)


ERROR_EMOTE_PERMISSION_DENIED = "You are not allowed to modify this emote."
ERROR_EMOTE_EMPTY_NAME = "Empty emote name."
ERROR_EMOTE_UNKNOWN_ACTION = "Unknown action specified."


@bp.route("/change_emote/<int:emote_id>", methods=("POST",))
@login_required
def change_emote(emote_id):
    return_url = request.headers.get("Referer", url_for(".emotes"))
    emote = Emote.query.get_or_404(emote_id)
    ajax = bool(request.args.get("ajax"))

    if emote.channel_id != g.user.id:
        return make_error(ERROR_EMOTE_PERMISSION_DENIED, return_url, 403)

    action = request.form.get("action")
    if action == "update":
        emote_name = request.form.get("name")

        # Empty emote name
        if not emote_name:
            return make_error(ERROR_EMOTE_EMPTY_NAME, return_url)

        # Check for existing emote name
        if db.session.query(exists().where(
                Emote.emote_name == emote_name
        )).scalar():
            return make_error(ERROR_EMOTE_EXISTS.format(emote_name),
                              return_url)

        emote.emote_name = emote_name
        if not ajax:
            flash("Successfully updated emote.", "info")
    elif action == "delete":
        try:
            os.remove(os.path.join(
                current_app.config["UPLOAD_FOLDER"],
                emote.emote_url
            ))
        except FileNotFoundError:
            pass

        db.session.delete(emote)
        if not ajax:
            flash("Successfully deleted emote.", "info")
    else:
        return make_error(ERROR_EMOTE_UNKNOWN_ACTION, return_url)

    db.session.commit()

    if request.args.get("ajax"):
        return "", 204
    else:
        return redirect(return_url)


# Registered in __init__.create_app
def serve_emote(filename):
    return send_from_directory(current_app.config["UPLOAD_FOLDER"], filename)
